#!/usr/bin/perl -w
#
#############################################################################
#
# File: install.pl
#
# Purpose: Installer for fwknop
#
#############################################################################
#
# $Id: install.pl,v 1.8 2004/07/08 21:42:27 mbr Exp $
#

use Cwd;
use File::Copy;
use File::Path;
use Getopt::Long;
use Sys::Hostname;
use strict;

#========================== config ===========================
my $libdir     = '/usr/lib/fwknop';
my $confdir    = '/etc/fwknop';
my $vardir     = '/var/log/fwknop';
my $varlibdir  = '/var/lib/fwknop';
my $piddir     = '/var/run/fwknop';
my $archivedir = "$confdir/archive";
my $bindir     = '/usr/bin';
my $sbindir    = '/usr/sbin';
my $initdir    = '/etc/init.d';

my $fwknop_fifo = "$varlibdir/fwknopfifo";

### system binaries
my $chkconfigCmd = '/sbin/chkconfig';
my $rcupdateCmd  = '/sbin/rc-update';  ### Gentoo
my $makeCmd      = '/usr/bin/make';
my $perlCmd      = '/usr/bin/perl';
my $gzipCmd      = '/bin/gzip';
my $killallCmd   = '/usr/bin/killall';
my $mknodCmd     = '/bin/mknod';
#======================== end config =========================

my $print_help = 0;
my $uninstall  = 0;
my $runlevel;

my %cmds = (
    'make'    => $makeCmd,
    'perl'    => $perlCmd,
    'gzip'    => $gzipCmd,
    'killall' => $killallCmd,
    'mknod'   => $mknodCmd
);

&usage(1) unless (GetOptions(
    'uninstall'   => \$uninstall,   ### Uninstall fwknop.
    'help'        => \$print_help   ### Display help.
));
&usage(0) if $print_help;

### check to make sure we are running as root
$< == 0 && $> == 0 or die "[*] You need to be root (or ",
    "equivalent UID 0 account) to install/uninstall fwknop!";

my $distro = &get_distro();

if ($distro eq 'redhat') {
    ### add chkconfig only if we are runing on a redhat distro
    $cmds{'chkconfig'} = $chkconfigCmd;
} elsif ($distro eq 'gentoo') {
    ### add rc-update if we are running on a gentoo distro
    $cmds{'rc-update'} = $rcupdateCmd;
}

### make sure the system binaries are where we expect
### them to be.
&check_commands();

my $hostname = hostname();

if (not $uninstall) {
    &install();
} else {
    &uninstall();
}
exit 0;
#======================= end main ==========================

sub install() {
    print "[+] Installing fwknop on $hostname\n";
    unless (-d '/usr/lib') {
        mkdir '/usr/lib' or die "[*] Could not mkdir /usr/lib: $!";
    }
    unless (-d '/var/run') {
        mkdir '/var/run' or die "[*] Could not mkdir /var/run: $!";
    }
    unless (-d '/var/log') {
        mkdir '/var/log' or die "[*] Could not mkdir /var/log: $!";
    }
    unless (-d $piddir) {
        mkdir $piddir or die "[*] Could not mkdir $piddir: $!";
    }
    unless (-d $varlibdir) {
        mkdir $varlibdir or die "[*] Could not mkdir $varlibdir: $!";
    }
    unless (-d $bindir) {
        mkdir $bindir or die "[*] Could not mkdir $bindir: $!";
    }
    unless (-d $sbindir) {
        mkdir $sbindir or die "[*] Could not mkdir $sbindir: $!";
    }

    ### config directory
    if (-d $confdir) {
        print "[-] Config directory $confdir already exists.\n";
    } else {
        ### Note that root will only be able to view files in
        ### /etc/fwknop since fwknop only needs to view fwknop.conf
        ### when being run as a daemon.
        print "[+] Creating config directory: $confdir\n";
        mkdir $confdir, 0500 or die "[*] Could not mkdir $confdir: $!";
    }

    if (-d $vardir) {
        print "[-] Cache directory $vardir already exists.\n";
    } else {
        print "[+] Creating cache directory: $vardir\n";
        mkdir $vardir, 0500 or die "[*] Could not mkdir $vardir: $!";
    }

    ### archive directory for previously install config files
    if (-d $archivedir) {
        print "[-] Archive directory $archivedir already exists.\n";
    } else {
        print "[+] Creating archive directory: $archivedir\n";
        mkdir $archivedir, 0500 or die "[*] Could not mkdir $archivedir: $!";
    }

    ### make our library directory (for Unix::Syslog)
    if (-d $libdir) {
        print "[-] Lib directory $libdir already exists.\n";
    } else {
        print "[+] Creating directory $libdir for ",
            "Unix::Syslog perl module.\n";
        mkdir $libdir, 0755 or die "[*] Could not mkdir $libdir: $!";
    }

    ### install Unix::Syslog
    chdir 'Unix-Syslog' or die "[*] Could not chdir to ",
        "Unix-Syslog: $!";
    unless (-e 'Makefile.PL' && -e 'Syslog.pm') {
        die "[*] Your source directory appears to be incomplete!  Syslog.pm ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
    system $cmds{'make'};
    system "$cmds{'make'} install";
    chdir '..';

    ### install Net::IPv4Addr
    chdir 'Net-IPv4Addr' or die "[*] Could not chdir to ",
        "Net-IPv4Addr: $!";
    unless (-e 'Makefile.PL') {
        die "[*] Your source directory appears to be incomplete!  Makefile.PL ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
    system $cmds{'make'};
    system "$cmds{'make'} install";
    chdir '..';

    ### install Crypt::Rijndael
    chdir 'Crypt-Rijndael' or die "[*] Could not chdir to ",
        "Crypt-Rijndael: $!";
    unless (-e 'Makefile.PL') {
        die "[*] Your source directory appears to be incomplete!   Makefile.PL ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
    system $cmds{'make'};
    system "$cmds{'make'} install";
    chdir '..';

    ### install Crypt::CBC
    chdir 'Crypt-CBC' or die "[*] Could not chdir to ",
        "Crypt-CBC: $!";
    unless (-e 'Makefile.PL') {
        die "[*] Your source directory appears to be incomplete!   Makefile.PL ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
    system $cmds{'make'};
    system "$cmds{'make'} install";
    chdir '..';

    ### install TermReadKey
    chdir 'TermReadKey' or die "[*] Could not chdir to ",
        "TermReadKey: $!";
    unless (-e 'Makefile.PL') {
        die "[*] Your source directory appears to be incomplete!   Makefile.PL ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
    system $cmds{'make'};
    system "$cmds{'make'} install";
    chdir '..';

    ### create the named pipe
    unless (-e $fwknop_fifo and -p $fwknop_fifo) {
        unlink $fwknop_fifo if -e $fwknop_fifo;
        print "[+] Creating named pipe $fwknop_fifo\n";
        my $created_pipe = 1;
        unless (((system "$cmds{'mknod'} -m 600 $fwknop_fifo p")>>8) == 0) {
            $created_pipe = 0;
        }
        unless (-e $fwknop_fifo and -p $fwknop_fifo) {
            $created_pipe = 0;
        }
        unless ($created_pipe) {
            die "[*] Could not create the named pipe \"$fwknop_fifo\"!\n",
                "[*] fwknop requires this file to exist!  Aborting install.\n";
        }
    }

    my $restarted_syslog = 0;
    if (-e '/etc/syslog.conf') {
        &append_fifo_syslog();
        if (((system "$cmds{'killall'} -HUP syslogd 2> /dev/null")>>8) == 0) {
            print "[+] Restarted syslog.\n";
            $restarted_syslog = 1;
        }
    }
    if (-e '/etc/syslog-ng/syslog-ng.conf') {
        &append_fifo_syslog_ng();
        if (((system "$cmds{'killall'} -HUP syslog-ng 2> /dev/null")>>8) == 0) {
            print "[+] Restarted syslog-ng.\n";
            $restarted_syslog = 1;
        }
    }

    unless ($restarted_syslog) {
        print "[*] Could not restart any syslog daemons.\n";
    }
    unless (-e "${vardir}/fwdata") {
        print "[+] Creating ${vardir}/fwdata file\n";
        open F, "> ${vardir}/fwdata" or die " ** Could not open ",
            "${vardir}/fwdata: $!";
        close F;
        chmod 0600, "${vardir}/fwdata";
        &perms_ownership("${vardir}/fwdata", 0600);
    }

    ### install man pages
    &install_manpage('knopmd.8');
    &install_manpage('knopwatchd.8');
    &install_manpage('fwknop.8');

    my $init_file = '';
    if ($distro eq 'redhat') {
        $init_file = 'fwknop-init.redhat';
    } elsif ($distro eq 'gentoo') {
        $init_file = 'fwknop-init.gentoo';
    } else {
        $init_file = 'fwknop-init.generic';
    }

    ### install all config and access files
    for my $file qw(fwknop.conf
            knopmd.conf
            knopwatchd.conf
            pf.os posf
            access.conf) {
        &archive("$confdir/$file");
        print "[+] Copying $file -> $confdir\n";
        copy $file, $confdir or
            die "[*] Could not cp $file to $confdir";
        if ($file eq 'knopwatchd.conf') {
            &set_hostname($file);
        }
        chmod 0600, "$confdir/$file" or
            die "[*] Could not chmod(600, $confdir/$file: $!";
        chown 0, 0, "$confdir/$file" or
            die "[*] Could not chown 0, 0, $confdir/$file: $!";
    }

    print "[+] Compiling knopmd, and knopwatchd daemons:";

    ### remove any previously compiled knopmd
    unlink 'knopmd' if -e 'knopmd';

    ### remove any previously compiled knopwatchd
    unlink 'knopwatchd' if -e 'knopwatchd';

    ### compile the C fwknop daemons
    system $cmds{'make'};

    unless (-e 'knopmd' and -e 'knopwatchd') {
        die "[*] Compilation failed.";
    }

    ### install fwknop daemons
    for my $daemon qw(knopmd knopwatchd) {
        print "[+] Copying $daemon -> $sbindir\n";
        copy $daemon, $sbindir or
            die "[*] Could not cp $daemon to $sbindir";
        chmod 0500, "$sbindir/$daemon" or
            die "[*] Could not chmod 500 $sbindir/$daemon";
    }

    print "[+] Copying fwknop -> $bindir\n";
    copy 'fwknop', $bindir or
        die "[*] Could not cp fwknop to $bindir";
    chmod 0755, "$bindir/fwknop" or
        die "[*] Could not chmod 755 $bindir/fwknop";

    if (-d $initdir) {
        print "[+] Copying init-scripts/$init_file -> ${initdir}/fwknop\n";
        copy "init-scripts/$init_file", "${initdir}/fwknop";
        &perms_ownership("${initdir}/fwknop", 0744);
        &enable_fwknop_at_boot($distro);
    }
    print "\n[+] fwknop has been installed!  To start in server mode, run\n",
        "    /etc/init.d/fwknop start\n";
    print "    Note: You will need to edit $confdir/access.conf for fwknop to\n",
        "    function properly in server mode.\n";

    return;
}

sub uninstall() {
    ### FIXME: need to pid check to see if fwknop is still running
    ### before we remove it.
    print "[+] Uninstalling fwknop\n";
    rmtree $confdir, 1, 1 if -d $confdir;
    rmtree $libdir, 1, 1 if -d $libdir;
    unlink "$bindir/fwknop" if -e "$bindir/fwknop";
    return;
}

### check paths to commands and attempt to correct if any are wrong.
sub check_commands() {
    my @path = qw(
        /bin
        /sbin
        /usr/bin
        /usr/sbin
        /usr/local/bin
        /usr/local/sbin
    );
    CMD: for my $cmd (keys %cmds) {
        unless (-x $cmds{$cmd}) {
            my $found = 0;
            PATH: for my $dir (@path) {
                if (-x "${dir}/${cmd}") {
                    $cmds{$cmd} = "${dir}/${cmd}";
                    $found = 1;
                    last PATH;
                }
            }
            unless ($found) {
                die "[*] Could not find $cmd anywhere!!!  ",
                    "Please edit the config section to include the path to ",
                    "$cmd.\n";
            }
        }
        unless (-x $cmds{$cmd}) {
            die "[*] $cmd is located at ",
                "$cmds{$cmd} but is not executable by uid: $<\n";
        }
    }
    return;
}

sub archive() {
    my $file = shift;
    return unless -e $file;
    my $curr_pwd = cwd();
    chdir $archivedir or die $!;
    my ($filename) = ($file =~ m|.*/(.*)|);
    my $base = "${filename}.old";
    for (my $i = 5; $i > 1; $i--) {  ### keep five copies of old config files
        my $j = $i - 1;
        unlink "${base}${i}.gz" if -e "${base}${i}.gz";
        move "${base}${j}.gz", "${base}${i}.gz" if -e "${base}${j}.gz";
    }
    print "[+] Archiving $file -> ${base}1\n";
    unlink "${base}1.gz" if -e "${base}1.gz";
    copy $file, "${base}1";   ### move $file into the archive directory
    system "$cmds{'gzip'} ${base}1";
    chdir $curr_pwd or die $!;
    return;
}

sub append_fifo_syslog() {
    print "[+] Modifying /etc/syslog.conf to write kern.info messages to\n";
    print "    $fwknop_fifo\n";
    unless (-e '/etc/syslog.conf.orig') {
        copy '/etc/syslog.conf', '/etc/syslog.conf.orig';
    }
    &archive('/etc/syslog.conf');
    open RS, '< /etc/syslog.conf' or
        die " **  Unable to open /etc/syslog.conf: $!\n";
    my @slines = <RS>;
    close RS;
    open SYSLOG, '> /etc/syslog.conf' or
        die " **  Unable to open /etc/syslog.conf: $!\n";
    for my $line (@slines) {
        unless ($line =~ /fwknopfifo/) {
            print SYSLOG $line;
        }
    }
    ### reinstate kernel logging to our named pipe
    print SYSLOG '### Send kern.info messages to fwknopfifo for ',
        "analysis by knopmd\n",
        "kern.info\t\t|$fwknop_fifo\n";
    close SYSLOG;
    return;
}

sub append_fifo_syslog_ng() {
    print "[+] Modifying /etc/syslog-ng/syslog-ng.conf to write kern.info ",
        "messages to\n";
    print "    $fwknop_fifo\n";
    unless (-e '/etc/syslog-ng/syslog-ng.conf.orig') {
        copy '/etc/syslog-ng/syslog-ng.conf',
            '/etc/syslog-ng/syslog-ng.conf.orig';
    }
    &archive('/etc/syslog-ng/syslog-ng.conf');
    open RS, '< /etc/syslog-ng/syslog-ng.conf' or
        die " **  Unable to open /etc/syslog-ng/syslog-ng.conf: $!\n";
    my @slines = <RS>;
    close RS;

    my $found_fifo = 0;
    for my $line (@slines) {
        $found_fifo = 1 if ($line =~ /fwknopfifo/);
    }

    unless ($found_fifo) {
        open SYSLOGNG, '>> /etc/syslog-ng/syslog-ng.conf' or
            die " ** Unable to open /etc/syslog-ng/syslog-ng.conf: $!\n";
        print SYSLOGNG "\n",
            "destination fwknoppipe { pipe(\"/var/lib/fwknop/fwknopfifo\"); };\n",
            "filter f_kerninfo { facility(kern) and level(info); };\n",
            "log { source(src); filter(f_kerninfo); destination(fwknoppipe); };\n";
        close SYSLOGNG;
    }
    return;
}

sub get_distro() {
    return 'gentoo' if -e '/etc/gentoo-release';
    if (-e '/etc/issue') {
        ### Red Hat Linux release 6.2 (Zoot)
        open ISSUE, '< /etc/issue' or
            die " ** Could not open /etc/issue: $!";
        my @lines = <ISSUE>;
        close ISSUE;
        for my $line (@lines) {
            chomp $line;
            return 'redhat' if $line =~ /red\s*hat/i;
        }
    }
    return 'NA';
}

sub install_manpage() {
    my $manpage = shift;

    ### default location to put man pages, but check with
    ### /etc/man.config
    my $mpath = '/usr/share/man/man8';
    if (-e '/etc/man.config') {
        ### prefer to install $manpage in /usr/local/man/man8 if
        ### this directory is configured in /etc/man.config
        open M, '< /etc/man.config' or
            die "[*] Could not open /etc/man.config: $!";
        my @lines = <M>;
        close M;
        ### prefer the path "/usr/share/man"
        my $found = 0;
        for my $line (@lines) {
            chomp $line;
            if ($line =~ m|^MANPATH\s+/usr/share/man|) {
                $found = 1;
                last;
            }
        }
        ### try to find "/usr/local/man" if we didn't find /usr/share/man
        unless ($found) {
            for my $line (@lines) {
                chomp $line;
                if ($line =~ m|^MANPATH\s+/usr/local/man|) {
                    $mpath = '/usr/local/man/man8';
                    $found = 1;
                    last;
                }
            }
        }
        ### if we still have not found one of the above man paths,
        ### just select the first one out of /etc/man.config
        unless ($found) {
            for my $line (@lines) {
                chomp $line;
                if ($line =~ m|^MANPATH\s+(\S+)|) {
                    $mpath = $1;
                    last;
                }
            }
        }
    }
    mkdir $mpath, 0755 unless -d $mpath;
    my $mfile = "${mpath}/${manpage}";
    print "[+] Installing $manpage man page at $mfile\n";
    copy $manpage, $mfile or die " ** Could not copy $manpage to ",
        "$mfile: $!";
    &perms_ownership($mfile, 0644);
    print "[+] Compressing manpage $mfile\n";
    ### remove the old one so gzip doesn't prompt us
    unlink "${mfile}.gz" if -e "${mfile}.gz";
    system "$cmds{'gzip'} $mfile";
    return;
}

sub enable_fwknop_at_boot() {
    my $distro = shift;
    my $ans = '';
    while ($ans ne 'y' && $ans ne 'n') {
        print "[+] Enable fwknop at boot time ([y]/n)?  ";
        $ans = <STDIN>;
        if ($ans eq "\n") {  ### allow the default of "y" to take over
            $ans = 'y';
        }
        chomp $ans;
    }
    if ($ans eq 'y') {
        if ($distro eq 'redhat') {
            system "$cmds{'chkconfig'} --add fwknop";
        } elsif ($distro eq 'gentoo') {
            system "$cmds{'rc-update'} add fwknop default";
        } else {  ### it is a non-redhat distro, try to
                  ### get the runlevel from /etc/inittab
            if ($runlevel) {
                ### the link already exists, so don't re-create it
                unless (-e "/etc/rc.d/rc${runlevel}.d/S99fwknop") {
                    symlink '/etc/rc.d/init.d/fwknop',
                        "/etc/rc.d/rc${runlevel}.d/S99fwknop";
                }
            } elsif (-e '/etc/inittab') {
                open I, '< /etc/inittab' or die " ** Could not open ",
                    "/etc/inittab: $!";
                my @ilines = <I>;
                close I;
                for my $line (@ilines) {
                    chomp $line;
                    if ($line =~ /^id\:(\d)\:initdefault/) {
                        $runlevel = $1;
                        last;
                    }
                }
                unless ($runlevel) {
                    print "[-]  Could not determine the runlevel.  Set ",
                        "the runlevel\nmanually in the config section of ",
                        "install.pl\n";
                    return;
                }
                ### the link already exists, so don't re-create it
                unless (-e "/etc/rc.d/rc${runlevel}.d/S99fwknop") {
                    symlink '/etc/rc.d/init.d/fwknop',
                        "/etc/rc.d/rc${runlevel}.d/S99fwknop";
                }
            } else {
                print "[-] /etc/inittab does not exist!  Set the ",
                    "runlevel\nmanually in the config section of ",
                    "install.pl.\n";
                return;
            }
        }
    }
    return;
}

sub set_hostname() {
    my $file = shift;
    if (-e $file) {
        open P, "< $file" or die "[*] Could not open $file: $!";
        my @lines = <P>;
        close P;
        ### replace the "HOSTNAME           CHANGE_ME" line
        open PH, "> $file" or die "[*] Could not open $file: $!";
        for my $line (@lines) {
            chomp $line;
            if ($line =~ /^\s*HOSTNAME(\s+)_?CHANGE.?ME_?/) {
                print PH "HOSTNAME${1}$hostname;\n";
            } else {
                print PH "$line\n";
            }
        }
        close PH;
    } else {
        die "[*] Your source directory appears to be incomplete!  $file ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    return;
}

sub perms_ownership() {
    my ($file, $perm_value) = @_;
    chmod $perm_value, $file or die " ** Could not ",
        "chmod($perm_value, $file): $!";
    ### root (maybe should take the group assignment out)
    chown 0, 0, $file or die " ** Could not chown 0,0,$file: $!";
    return;
}

sub usage() {
    my $exit_status = shift;
    print "usage\n";
    exit $exit_status;
}
